# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'tikka/slots/pyqt/resources/gui/windows/welcome.ui'
#
# Created by: PyQt5 UI code generator 5.15.10
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_WelcomeDialog(object):
    def setupUi(self, WelcomeDialog):
        WelcomeDialog.setObjectName("WelcomeDialog")
        WelcomeDialog.resize(505, 672)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(WelcomeDialog.sizePolicy().hasHeightForWidth())
        WelcomeDialog.setSizePolicy(sizePolicy)
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(WelcomeDialog)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.label = QtWidgets.QLabel(WelcomeDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label.sizePolicy().hasHeightForWidth())
        self.label.setSizePolicy(sizePolicy)
        self.label.setPixmap(QtGui.QPixmap(":/icons/logo"))
        self.label.setAlignment(QtCore.Qt.AlignCenter)
        self.label.setWordWrap(False)
        self.label.setObjectName("label")
        self.verticalLayout_2.addWidget(self.label)
        self.label_3 = QtWidgets.QLabel(WelcomeDialog)
        self.label_3.setAlignment(QtCore.Qt.AlignCenter)
        self.label_3.setObjectName("label_3")
        self.verticalLayout_2.addWidget(self.label_3)
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setContentsMargins(-1, 20, -1, 20)
        self.verticalLayout.setObjectName("verticalLayout")
        self.importV2Button = QtWidgets.QPushButton(WelcomeDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.importV2Button.sizePolicy().hasHeightForWidth())
        self.importV2Button.setSizePolicy(sizePolicy)
        self.importV2Button.setObjectName("importV2Button")
        self.verticalLayout.addWidget(self.importV2Button, 0, QtCore.Qt.AlignHCenter)
        self.importV1Button = QtWidgets.QPushButton(WelcomeDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.importV1Button.sizePolicy().hasHeightForWidth())
        self.importV1Button.setSizePolicy(sizePolicy)
        self.importV1Button.setObjectName("importV1Button")
        self.verticalLayout.addWidget(self.importV1Button, 0, QtCore.Qt.AlignHCenter)
        self.createButton = QtWidgets.QPushButton(WelcomeDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.createButton.sizePolicy().hasHeightForWidth())
        self.createButton.setSizePolicy(sizePolicy)
        self.createButton.setObjectName("createButton")
        self.verticalLayout.addWidget(self.createButton, 0, QtCore.Qt.AlignHCenter)
        self.verticalLayout_2.addLayout(self.verticalLayout)
        self.continueButton = QtWidgets.QPushButton(WelcomeDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.continueButton.sizePolicy().hasHeightForWidth())
        self.continueButton.setSizePolicy(sizePolicy)
        self.continueButton.setObjectName("continueButton")
        self.verticalLayout_2.addWidget(self.continueButton, 0, QtCore.Qt.AlignHCenter)

        self.retranslateUi(WelcomeDialog)
        QtCore.QMetaObject.connectSlotsByName(WelcomeDialog)
        WelcomeDialog.setTabOrder(self.createButton, self.continueButton)

    def retranslateUi(self, WelcomeDialog):
        #  _translate = QtCore.QCoreApplication.translate
        WelcomeDialog.setWindowTitle(_("Welcome"))
        self.label_3.setText(_("Welcome in Tikka !\n"
"\n"
"To get started, you can import an existing account or create a new one."))
        self.importV2Button.setText(_("Import account with mnemonic"))
        self.importV1Button.setText(_("Import V1 account with credentials"))
        self.createButton.setText(_("Create a new account"))
        self.continueButton.setText(_("Continue"))
import tikka.slots.pyqt.resources.icons.index_rc
