# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'tikka/slots/pyqt/resources/gui/windows/account_create.ui'
#
# Created by: PyQt5 UI code generator 5.15.10
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_AccountCreateDialog(object):
    def setupUi(self, AccountCreateDialog):
        AccountCreateDialog.setObjectName("AccountCreateDialog")
        AccountCreateDialog.resize(924, 497)
        self.verticalLayout = QtWidgets.QVBoxLayout(AccountCreateDialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.imagelabel = QtWidgets.QLabel(AccountCreateDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.imagelabel.sizePolicy().hasHeightForWidth())
        self.imagelabel.setSizePolicy(sizePolicy)
        self.imagelabel.setMinimumSize(QtCore.QSize(0, 0))
        self.imagelabel.setMaximumSize(QtCore.QSize(200, 200))
        self.imagelabel.setSizeIncrement(QtCore.QSize(0, 0))
        self.imagelabel.setBaseSize(QtCore.QSize(0, 0))
        self.imagelabel.setText("")
        self.imagelabel.setPixmap(QtGui.QPixmap(":/images/safe_background"))
        self.imagelabel.setScaledContents(True)
        self.imagelabel.setAlignment(QtCore.Qt.AlignCenter)
        self.imagelabel.setObjectName("imagelabel")
        self.verticalLayout.addWidget(self.imagelabel, 0, QtCore.Qt.AlignHCenter)
        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.setObjectName("gridLayout")
        self.addressLabel = QtWidgets.QLabel(AccountCreateDialog)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.addressLabel.setFont(font)
        self.addressLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.addressLabel.setObjectName("addressLabel")
        self.gridLayout.addWidget(self.addressLabel, 5, 0, 1, 1)
        self.nameLabel = QtWidgets.QLabel(AccountCreateDialog)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.nameLabel.setFont(font)
        self.nameLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.nameLabel.setObjectName("nameLabel")
        self.gridLayout.addWidget(self.nameLabel, 6, 0, 1, 1)
        self.mnemonicLineEdit = QtWidgets.QLineEdit(AccountCreateDialog)
        self.mnemonicLineEdit.setReadOnly(True)
        self.mnemonicLineEdit.setObjectName("mnemonicLineEdit")
        self.gridLayout.addWidget(self.mnemonicLineEdit, 0, 1, 1, 1)
        self.mnemonicLabel = QtWidgets.QLabel(AccountCreateDialog)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.mnemonicLabel.setFont(font)
        self.mnemonicLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.mnemonicLabel.setObjectName("mnemonicLabel")
        self.gridLayout.addWidget(self.mnemonicLabel, 0, 0, 1, 1)
        self.nameLineEdit = QtWidgets.QLineEdit(AccountCreateDialog)
        self.nameLineEdit.setObjectName("nameLineEdit")
        self.gridLayout.addWidget(self.nameLineEdit, 6, 1, 1, 1)
        self.addressValueLabel = QtWidgets.QLabel(AccountCreateDialog)
        self.addressValueLabel.setText("")
        self.addressValueLabel.setTextInteractionFlags(QtCore.Qt.LinksAccessibleByMouse|QtCore.Qt.TextSelectableByKeyboard|QtCore.Qt.TextSelectableByMouse)
        self.addressValueLabel.setObjectName("addressValueLabel")
        self.gridLayout.addWidget(self.addressValueLabel, 5, 1, 1, 1)
        self.mnemonicLanguageComboBox = QtWidgets.QComboBox(AccountCreateDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.mnemonicLanguageComboBox.sizePolicy().hasHeightForWidth())
        self.mnemonicLanguageComboBox.setSizePolicy(sizePolicy)
        self.mnemonicLanguageComboBox.setObjectName("mnemonicLanguageComboBox")
        self.gridLayout.addWidget(self.mnemonicLanguageComboBox, 2, 1, 1, 1)
        self.mnemonicLanguageLabel = QtWidgets.QLabel(AccountCreateDialog)
        font = QtGui.QFont()
        font.setBold(True)
        font.setItalic(False)
        font.setWeight(75)
        self.mnemonicLanguageLabel.setFont(font)
        self.mnemonicLanguageLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.mnemonicLanguageLabel.setObjectName("mnemonicLanguageLabel")
        self.gridLayout.addWidget(self.mnemonicLanguageLabel, 2, 0, 1, 1)
        self.verticalLayout.addLayout(self.gridLayout)
        self.changeButton = QtWidgets.QPushButton(AccountCreateDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(7)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.changeButton.sizePolicy().hasHeightForWidth())
        self.changeButton.setSizePolicy(sizePolicy)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/icons/refresh_icon"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.changeButton.setIcon(icon)
        self.changeButton.setObjectName("changeButton")
        self.verticalLayout.addWidget(self.changeButton, 0, QtCore.Qt.AlignHCenter)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.saveMnemonicCheckBox = QtWidgets.QCheckBox(AccountCreateDialog)
        self.saveMnemonicCheckBox.setObjectName("saveMnemonicCheckBox")
        self.horizontalLayout_2.addWidget(self.saveMnemonicCheckBox)
        self.verticalLayout.addLayout(self.horizontalLayout_2)
        self.passwordFrame = QtWidgets.QFrame(AccountCreateDialog)
        self.passwordFrame.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.passwordFrame.setFrameShadow(QtWidgets.QFrame.Raised)
        self.passwordFrame.setObjectName("passwordFrame")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.passwordFrame)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.passwordLabel = QtWidgets.QLabel(self.passwordFrame)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.passwordLabel.sizePolicy().hasHeightForWidth())
        self.passwordLabel.setSizePolicy(sizePolicy)
        self.passwordLabel.setSizeIncrement(QtCore.QSize(0, 0))
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.passwordLabel.setFont(font)
        self.passwordLabel.setAlignment(QtCore.Qt.AlignLeading|QtCore.Qt.AlignLeft|QtCore.Qt.AlignVCenter)
        self.passwordLabel.setObjectName("passwordLabel")
        self.horizontalLayout.addWidget(self.passwordLabel)
        self.passwordLineEdit = QtWidgets.QLineEdit(self.passwordFrame)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.passwordLineEdit.sizePolicy().hasHeightForWidth())
        self.passwordLineEdit.setSizePolicy(sizePolicy)
        self.passwordLineEdit.setMaxLength(20)
        self.passwordLineEdit.setAlignment(QtCore.Qt.AlignCenter)
        self.passwordLineEdit.setReadOnly(True)
        self.passwordLineEdit.setObjectName("passwordLineEdit")
        self.horizontalLayout.addWidget(self.passwordLineEdit)
        self.passwordChangeButton = QtWidgets.QPushButton(self.passwordFrame)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.passwordChangeButton.sizePolicy().hasHeightForWidth())
        self.passwordChangeButton.setSizePolicy(sizePolicy)
        self.passwordChangeButton.setIcon(icon)
        self.passwordChangeButton.setObjectName("passwordChangeButton")
        self.horizontalLayout.addWidget(self.passwordChangeButton)
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem1)
        self.verticalLayout_2.addLayout(self.horizontalLayout)
        self.verticalLayout.addWidget(self.passwordFrame)
        self.buttonBox = QtWidgets.QDialogButtonBox(AccountCreateDialog)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.verticalLayout.addWidget(self.buttonBox, 0, QtCore.Qt.AlignHCenter)
        spacerItem2 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout.addItem(spacerItem2)

        self.retranslateUi(AccountCreateDialog)
        self.buttonBox.accepted.connect(AccountCreateDialog.accept) # type: ignore
        self.buttonBox.rejected.connect(AccountCreateDialog.reject) # type: ignore
        QtCore.QMetaObject.connectSlotsByName(AccountCreateDialog)
        AccountCreateDialog.setTabOrder(self.mnemonicLineEdit, self.nameLineEdit)
        AccountCreateDialog.setTabOrder(self.nameLineEdit, self.changeButton)
        AccountCreateDialog.setTabOrder(self.changeButton, self.saveMnemonicCheckBox)
        AccountCreateDialog.setTabOrder(self.saveMnemonicCheckBox, self.passwordLineEdit)
        AccountCreateDialog.setTabOrder(self.passwordLineEdit, self.passwordChangeButton)

    def retranslateUi(self, AccountCreateDialog):
        #  _translate = QtCore.QCoreApplication.translate
        AccountCreateDialog.setWindowTitle(_("Create account"))
        self.addressLabel.setText(_("Address"))
        self.nameLabel.setText(_("Name"))
        self.mnemonicLabel.setText(_("Mnemonic"))
        self.mnemonicLanguageLabel.setText(_("Mnemonic language"))
        self.changeButton.setText(_("Change"))
        self.saveMnemonicCheckBox.setText(_("I have saved my mnemonic phrase in a safety place"))
        self.passwordLabel.setText(_("Password"))
        self.passwordChangeButton.setText(_("Change"))
import tikka.slots.pyqt.resources.icons.index_rc
